// BB-...  
//     ,    
//
//        ,      ,
//        
//             ( )
//          ,     -   
 

#include<stdio.h>
#include<stdlib.h>

//! 
class BB_tree{
	//!  
	struct node_st {
		node_st *p1;	//!<  
		node_st *p2;	//!<  
		int value;		//!<  
		int count;		//!<     (    )
	//
		node_st(int v):p1(0),p2(0),count(1),value(v) {}		// 
		int Count() { return this?count:0;}					//  
		int Dir() { return this?p1->Count()-p2->Count():0;}	//   (    Eval)
	};
	node_st *tree_root;	//!<  
	int Eval(node_st *node);					//!<   
	bool Insert(node_st **node,int value);		//!<   ( )
	bool Remove(node_st **node,int value);		//!<  ( )
	void GetMin(node_st **root,node_st **res);	//!<    ( )
	void Balance(node_st**);					//!<  
	node_st *Rotate12(node_st*);				//!<  
	node_st *Rotate21(node_st*);				//!< 
	void Show(node_st*,int d);					//!<  
public:
	BB_tree();
	~BB_tree();
	void Show();				//!< 
	void Insert(int value);		//!<  
	void Remove(int value);		//!<  
	bool Find(int value);		//!<  
};


void BB_tree::Show()
{
	Show(tree_root,0);
}

void BB_tree::Show(node_st *root,int d)
{
	if(!root) return;
	int b=Eval(root);
	const char *info=""; if(b<0) info="[-]"; if(b>0) info="[+]"; //        Eval 
	Show(root->p1,d+1);
	printf("%*s%d: %d%s\n",d*2,"",root->value,root->count,info);
	Show(root->p2,d+1);
}


BB_tree::BB_tree()
{
	tree_root=0; 
}

BB_tree::~BB_tree()
{
	//   
	while(tree_root) {
		while(tree_root->p2) tree_root=Rotate21(tree_root);
		node_st *t=tree_root; tree_root=tree_root->p1;
		delete t;
	}
}


inline void BB_tree::Insert(int value)
{
	Insert(&tree_root,value);
}

inline void BB_tree::Remove(int value)
{
	Remove(&tree_root,value);
}

bool BB_tree::Find(int value)
{
	node_st *node=tree_root;
	while(node) {
		if(node->value==value) return true;
		node=node->value>value?node->p1:node->p2;
	}
	return false;
}


//   ,    BB- (   )
int BB_tree::Eval(node_st *node)
{
	int c1=node->p1->Count();
	int c2=node->p2->Count();
	if(c1/3>c2/2) return 1;
	if(c2/3>c1/2) return -1;
	return 0;
/*
    
-      return node->Dir() -         
- (c1/2>c2) (c1/2>c2) -  AVL 
- (c1/3>c2) (c1/3>c2) -    ,    ,     
- (c1/8>c2) (c1/8>c2) -       ,        
-  ,   ,   -....
*/
}


void BB_tree::Balance(node_st **root)
{
	node_st *node=*root;
	int b=Eval(node);
	if(b<0) {
		if(node->p2->Dir()>0) node->p2=Rotate12(node->p2);
		*root=Rotate21(node);
	}
	else if(b>0) {
		if(node->p1->Dir()<0) node->p1=Rotate21(node->p1);
		*root=Rotate12(node);
	}
}

bool BB_tree::Insert(node_st **root,int value)
{
	bool res=false;
	node_st *node=*root;
	if(!node) {
		*root=new node_st(value);
		return true;
	}
	if(value<node->value) res=Insert(&node->p1,value); else
	if(value>node->value) res=Insert(&node->p2,value);
	if(res) {
		node->count++;
		Balance(root);
	}
	return res;
}

BB_tree::node_st *BB_tree::Rotate12(node_st *node)
{
	node_st *p1=node->p1;
	node_st *p12=p1->p2;
	p1->p2=node;  
	node->p1=p12; 
	p1->count=node->count;
	node->count=1+node->p1->Count()+node->p2->Count();
	return p1;
}

BB_tree::node_st *BB_tree::Rotate21(node_st *node)
{
	node_st *p2=node->p2;
	node_st *p21=p2->p1;
	p2->p1=node;  
	node->p2=p21; 
	p2->count=node->count;
	node->count=1+node->p1->Count()+node->p2->Count();
	return p2;
}

// \retval res     
void BB_tree::GetMin(node_st **root,node_st **res)
{
	node_st *node=*root;
	if(node->p1) {
		GetMin(&node->p1,res);
		return;
	}
	*res=node;
	*root=node->p2;
}


bool BB_tree::Remove(node_st **root,int value)
{
	bool res=false;
	node_st *node=*root;
	if(!node) return res;
	if(node->value<value) {
		res=Remove(&node->p2,value);
	}else if(node->value>value) {
		res=Remove(&node->p1,value);
	}else {							//   .  
		res=true;
		if(!node->p2) {				//     , 
			*root=node->p1; delete node;
			return res;
		}
		GetMin(&node->p2,root);	//  ,     
		(*root)->count=node->count;		//     ,  
		(*root)->p1   =node->p1;
		(*root)->p2   =node->p2;
		delete node;
	} 
	if(res) {
		node->count--;
		Balance(root);
	}
	return res;
}




int main()
{
	int n,m;
	BB_tree tree;
	for(n=0; n<30; n++)	{
		m=rand()%30;
		printf("[ins: %d]\n",m);
		tree.Insert(m);
	}
	tree.Show();
	system("pause");
	for( n=0; n<30; n++)	{
		m=rand()%30; 
		printf("[del: %d]\n",m);
		tree.Remove(m);
	}
	tree.Show();
	system("pause");
	tree.Find(3);
	return 0;
}


